/*
    Copyright (c) 2026 Christian Schoenebeck

    This file is part of "gigedit" and released under the terms of the
    GNU General Public License version 2.
*/

#include "GEColor.h"

GEColor::GEColor(Glib::ustring bright, Glib::ustring dark) {
    m_bright = bright;
    m_dark = dark;
}

GEColor::operator Gdk::RGBA() const {
    return Gdk::RGBA(isDarkMode() ? m_dark : m_bright);
}

GEColor::operator Glib::ustring() const {
    return isDarkMode() ? m_dark : m_bright;
}


namespace GEColors {

const GEColor bronze("#c4950c", "#f2be00");

const GEColor comboTextColor("black", "white");
const GEColor multipleValuesColor(bronze);

} // namespace GEColors
