/*
    Copyright (c) 2026 Christian Schoenebeck

    This file is part of "gigedit" and released under the terms of the
    GNU General Public License version 2.
*/

#ifndef GIGEDIT_DRAW_H
#define GIGEDIT_DRAW_H

#include "compat.h"
#include <gtkmm/drawingarea.h>

void geDrawBubble(const Cairo::RefPtr<Cairo::Context>& cr,
                  double x, double y, double w, double h, double cornerRadius);

#endif // GIGEDIT_DRAW_H
